import os
from typing import Tuple

import sh


def get_cmd_args(cmd: sh.Command):
    args = [cmd._path]
    if cmd._partial:
        args.extend(cmd._partial_baked_args)
    return [e.decode('utf-8') for e in args]


def compile_shargs(*args, **kwargs) -> Tuple[list, dict]:
    shcmd = sh.Command('/bin/ls').bake(*args, **kwargs)
    return (
        [a.decode('utf-8') for a in shcmd._partial_baked_args],
        {'_' + k: i for k, i in shcmd._partial_call_args.items()}
    )


def source(path: str, cmd=None, _cwd='.', _env=os.environ) -> dict:
    ret = {}
    if cmd is None:
        envs = sh.bash('-c', 'source ' + path + ' >/dev/null && env', _cwd=_cwd, _env=_env)
    else:
        envs = sh.bash('-c', 'source ' + path + ' >/dev/null && ' + cmd + ' >/dev/null && env', _cwd=_cwd, _env=_env)
    for env in envs.splitlines():
        env: str
        var, val = env.split('=', 1)
        ret[var] = val
    return ret



import os
import hashlib
from typing import Union, List

import sh

from .fileutil import find_recursive
from .mputil import MpUtil


def hash_str(string: str) -> str:
    return hashlib.sha1(string.encode('utf-8')).hexdigest()


def hash_path(path: str) -> Union[str, None]:
    if not os.path.exists(path):
        return None

    files: List[str] = find_recursive(path, type='file')

    if not files:
        return None

#    return hash_str(sh.sha1sum(files))
    h = hashlib.sha1()
    for file in files:
        h.update(bytes(os.path.getmtime(file).hex().encode('ascii')))
    return h.hexdigest()


class FCache:
    cache_dir: str

    def __init__(self, cache_dir: str):
        self.cache_dir = os.path.realpath(cache_dir)

    def __get_cache(self, path: str) -> str:
        return os.path.join(self.cache_dir, hash_str(path))

    def update(self, src_path: str, value_path: str) -> bool:
        return False

        src_path = os.path.realpath(src_path)
        if not os.path.exists(src_path):
            return False

        value_path = os.path.realpath(value_path)
        if not os.path.exists(value_path):
            return False

        cache_path = self.__get_cache(src_path)

        if not os.path.isdir(cache_path):
            mkdir(cache_path)

        sign = hash_path(src_path)

        with open(os.path.join(cache_path, 'sign'), 'w+') as f:
            f.write(sign)

        overwrite(value_path, os.path.join(cache_path, 'value'))

        return True

    def get(self, src_path: str, dst_path: str) -> bool:
        return False

        src_path = os.path.realpath(src_path)
        if not os.path.exists(src_path):
            # print('miss: ' + src_path + ' to ' + dst_path)
            return False

        dst_path = os.path.realpath(dst_path)

        cache_path = self.__get_cache(src_path)

        if not os.path.isdir(cache_path):
            # print('miss: ' + src_path + ' to ' + dst_path)
            return False

        curr_sign = hash_path(src_path)

        with open(os.path.join(cache_path, 'sign'), 'r') as f:
            sign = f.read()

        if curr_sign != sign:
            # print('miss: ' + src_path + ' to ' + dst_path)
            return False

        overwrite(os.path.join(cache_path, 'value'), dst_path)

        return True



import getpass
import os
import sys
import tempfile
from os import path as _p
from typing import Union, List
# from deprecation import deprecated

import sh


def __build(build_context, fg=False, drop_priv=False):
    dockerfile = None

    if drop_priv:
        username = getpass.getuser()
        uid = os.getuid()
        gid = os.getgid()
        home = f'/home/{username}'
        groupname = f'$(getent group {gid} | cut -d: -f1)'

        with open(_p.join(build_context, 'Dockerfile'), 'r') as f:
            dockerfile = f.read()
        dockerfile += (
            f'\n'
            f'SHELL ["/bin/bash", "-c"]'
            f'\n'
            f'RUN if [ "$(id -u {username} > /dev/null 2>&1; echo $?)" == 0 ]; then userdel {username}; fi'
            f' && groupadd -g {gid} {username}'
            f' ;  useradd -l -u {uid} -g {groupname} {username}'
            f' && install -d -m 0755 -o {username} -g {groupname} {home}'
            f'\n'
            f'SHELL ["/bin/sh", "-c"]'
            f'\n'
            f'ENV HOME={home}'
            f'\n'
            f'ENV USER={username}'
            f'\n'
            f'USER {uid}:{gid}'
        )

    tmpdir = tempfile.TemporaryDirectory()

    iidfile = _p.join(tmpdir.name, '__iid')
    try:
        sh.docker.build('.',
                        f='-' if bool(drop_priv) else False,
                        iidfile=iidfile,
                        _in=dockerfile if bool(drop_priv) else None,
                        _cwd=build_context,
                        _err_to_out=bool(fg),
                        _out=sys.stdout if bool(fg) else None,
                        )
    except Exception:
        raise Exception(f'failed to build a docker image with build context at {build_context}')

    with open(iidfile, 'r') as f:
        return f.read()


def docker_sh(
        docker_context: str,
        root: bool = False,
        verbose: bool = False,
        volumes: Union[str, List[str]] = None,
        auto_remove: bool = True,
) -> sh.Command:

    if (not docker_context) or (not _p.isdir(docker_context)):
        raise Exception('proper docker_context directory must be supplied')

    if verbose:
        print('Building a docker image...')
    image_id = __build(docker_context, fg=verbose, drop_priv=not root)
    if not image_id:
        raise Exception('failed to build image')

    if root:
        home = '/root'
    else:
        username = getpass.getuser()
        home = f'/home/{username}'

    if not volumes:
        v_opts = []
    elif isinstance(volumes, str):
        v_opts = [f'-v={volumes}']
    elif isinstance(volumes, list):
        v_opts = [*map(lambda o: f'-v={o}', volumes)]
    else:
        raise Exception

    run = sh.docker.run.bake(
        *v_opts,
        '-d=false',
        i=True,
        rm=bool(auto_remove),  # Automatically remove the container when it exits
        workdir=home,
    )

    return run.bake(image_id)
